// Filename: XmlTester.cpp
// S.Chan, Jul 17, 2005

#include "stdafx.h"
#include "XmlTester.h"

#define MAX_BLANKS    1000

using namespace std;

// Helper function for implementing parsing delay.
// Technique from Chapter 17 of "Programming Windows with MFC,
// Second Edition", by Jeff Prosise (Microsoft Press).
static BOOL PumpWindowsMessages()
{
    MSG msg;
    
    int maxPumpCount = 10;
    
    // Dispatch any waiting messages.
    while ( maxPumpCount-- > 0 && PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE) )
    {
        if ( !AfxGetApp()->PumpMessage() ) 
        {
            PostQuitMessage(0);
            return FALSE;
        }
    }
    
    // Simulate the MFC's idle processing mechanism.
    int idle = 0;
    while ( true )
    {
        if ( !AfxGetApp()->OnIdle(idle) )
            break;
        ++idle;
    }

    return TRUE;
}

// CXmlTester //////////////////////////////////////////////////////////////

CXmlTester::CXmlTester()
{
    m_resultsWnd = NULL;
    m_parsingDelay = 0;
    m_abort = false;
    m_enableSchemaValidation = false;
    m_enableSchemaLocation = true;
    m_enableExhaustiveErrors = false;

    // Create XML parser and attach ourself as
    // the XML events/element handler.
    m_xmlParser = new CXmlParser;
    m_xmlParser->AttachElementHandler(this);
}

CXmlTester::~CXmlTester()
{
    delete m_xmlParser;
}

void CXmlTester::OnXmlStartElement(const CXmlElement& xmlElement)
{
    // Update results window.
    if ( m_resultsWnd != NULL )
    {
        m_resultsWnd->InsertString(-1, xmlElement.ToString().c_str());
    }

    // Check if we need to delay the parsing inbetween start elements.
    if ( m_parsingDelay > 0 )
    {
        // By pumping messages, we avoid freezing the GUI.
        // This gives the appearance of being multithreaded.
        PumpWindowsMessages();

        // Sleep for a duration in milliseconds.
        ::Sleep(m_parsingDelay);
    }
}

void CXmlTester::OnXmlElementData(const std::string& elementData, int depth)
{
    // Determine amount of leading whitespace. For each
    // depth level, we indent 4 blankspaces.
    char szBlanks[MAX_BLANKS];
    int numBlanks = min(MAX_BLANKS-1, 4 * depth);
    numBlanks = max(0, numBlanks);
    sprintf(szBlanks, "%*s", numBlanks, "");

    // Update results window.
    if ( m_resultsWnd != NULL )
    {
        CString strElem = szBlanks;
        strElem += CString(elementData.c_str());
        m_resultsWnd->InsertString(-1, strElem);
    }
}

void CXmlTester::OnXmlEndElement(const CXmlElement& xmlElement)
{
    // Update results window.
    if ( m_resultsWnd != NULL )
    {
        m_resultsWnd->InsertString(-1, xmlElement.ToString().c_str());
    }
}

void CXmlTester::OnXmlError(int line, int column, const std::string& errorText, unsigned long errorCode)
{
    // Update results window.
    if ( m_resultsWnd != NULL )
    {
        m_resultsWnd->InsertString(-1, " ");

        CString msg1 = errorText.c_str();
        msg1.TrimRight();

        CString msg2;
        msg2.Format("Log: line %d, column %d: errorCode=%d (0x%x)", line, column, errorCode, errorCode);
        m_resultsWnd->InsertString(-1, msg2);

        DisplayParseError(errorText);
    }
}

bool CXmlTester::OnXmlAbortParse(const CXmlElement& xmlElement)
{
    return m_abort;
}

void CXmlTester::AttachResultsWnd(CListBox* resultsWnd)
{
    m_resultsWnd = resultsWnd;
}

void CXmlTester::DetachResultsWnd()
{
    m_resultsWnd = NULL;
}

void CXmlTester::SetParsingDelay(int parsingDelay)
{
    m_parsingDelay = parsingDelay;
}

void CXmlTester::SetAbort(bool abort)
{
    m_abort = abort;
}

void CXmlTester::EnableSchemaValidation(bool enable)
{
    m_enableSchemaValidation = enable;
}

void CXmlTester::EnableSchemaLocation(bool enable)
{
    m_enableSchemaLocation = enable;
}

void CXmlTester::EnableExhaustiveErrors(bool enable)
{
    m_enableExhaustiveErrors = enable;
}

bool CXmlTester::AddValidationSchema(const std::string& namespaceURI, const std::string& xsdPath)
{
    if ( !m_xmlParser->IsReady() )
        return false;

    // Log initial message.
    if ( m_resultsWnd != NULL )
    {
        CString msg;
        msg.Format("Log: Adding schema: namespace URI=\"%s\",", namespaceURI.c_str());
        m_resultsWnd->InsertString(-1, msg);

        msg.Format("Log: with XSD path: %s", xsdPath.c_str());
        m_resultsWnd->InsertString(-1, msg);
    }

    // Add the validation schema to the parser.
    bool result = m_xmlParser->AddValidationSchema(namespaceURI, xsdPath);

    // Log a result message.
    if ( m_resultsWnd != NULL )
    {
        if ( result )
            m_resultsWnd->InsertString(-1, "Log: Add schema succeeded.");
        else
            m_resultsWnd->InsertString(-1, "Log: Add schema failed.");
        
        m_resultsWnd->InsertString(-1, " ");
    }

    return result;
}

bool CXmlTester::RemoveValidationSchema(const std::string& namespaceURI)
{
    if ( !m_xmlParser->IsReady() )
        return false;

    // Log initial message.
    if ( m_resultsWnd != NULL )
    {
        CString msg;
        msg.Format("Log: Removing schema: namespace URI=\"%s\"", namespaceURI.c_str());
        m_resultsWnd->InsertString(-1, msg);
    }

    // Remove the validation schema from the parser.
    bool result = m_xmlParser->RemoveValidationSchema(namespaceURI);

    // Log a result message.
    if ( m_resultsWnd != NULL )
    {
        if ( result )
            m_resultsWnd->InsertString(-1, "Log: Remove schema succeeded.");
        else
            m_resultsWnd->InsertString(-1, "Log: Remove schema failed (may already have been removed).");
        
        m_resultsWnd->InsertString(-1, " ");
    }

    return result;
}

bool CXmlTester::ParseXml(const std::string& xmlPath)
{
    if ( !m_xmlParser->IsReady() )
        return false;

    // Log initial message.
    if ( m_resultsWnd != NULL )
    {
        CString msg;
        msg.Format("Log: Parsing %s...", xmlPath.c_str());
        m_resultsWnd->InsertString(-1, msg);
        m_resultsWnd->InsertString(-1, " ");
    }

    // Set parser features/options.
    m_xmlParser->SetParserFeature("schema-validation", m_enableSchemaValidation);
    m_xmlParser->SetParserFeature("use-schema-location", m_enableSchemaLocation);
    m_xmlParser->SetParserFeature("exhaustive-errors", m_enableExhaustiveErrors);

    // Do the parsing.
    bool result = m_xmlParser->Parse(xmlPath);

    // Log a result message.
    if ( result )
    {
        m_resultsWnd->InsertString(-1, " ");
        m_resultsWnd->InsertString(-1, "Log: Parsing completed successfully.");
    }
    if ( m_abort )
    {
        m_resultsWnd->InsertString(-1, " ");
        m_resultsWnd->InsertString(-1, "Log: Parsing aborted by user.");
    }

    return result;
}

void CXmlTester::DisplayParseError(const std::string& errorText)
{
    if ( m_resultsWnd == NULL )
        return;
    if ( errorText.empty() )
        return;

    char szText[2000];
    strncpy(szText, errorText.c_str(), 1999);
    szText[1999] = 0;

    char* pszToken = strtok(szText, "\r\n");
    if ( pszToken != NULL )
    {
        CString msg;
        msg.Format("Log: %s", pszToken);
        m_resultsWnd->InsertString(-1, msg);

        while( true )
        {
            pszToken = strtok(NULL, "\r\n");
            if ( pszToken == NULL )
                break;

            msg.Format("Log: %s", pszToken);
            m_resultsWnd->InsertString(-1, msg);
        }
    }
}

// END
